import numpy as np


def ohe_to_str(
    ohe: np.ndarray,
    nucs: list[str] = ["A", "C", "G", "T", "N"]
) -> list[str]:
    """Convert OHE sequence to string representation.

    Args:
        ohe: One hot encoded sequence to convert.
        nucs: List of nucleotides corresponding to OHE position.

    Returns:
        List of string tokens representing nucleotides.
    """
    indices = np.where(ohe.sum(axis=-1) == 0, 4, np.argmax(ohe, axis=-1))
    sequences = ["".join(nucs[i] for i in row) for row in indices]
    sequences = [seq.rstrip("N") for seq in sequences]
    return sequences


def str_to_ohe(
    sequence: str,
    nucs: list[str] = ["A", "C", "G", "T"]
) -> np.ndarray:
    """Convert sequence to OHE.

    Args:
        sequence: Sequence to convert.
        nucs: Nucleotides corresponding to their one hot position.

    Returns:
        One hot encoded sequence.
    """
    mapping = {nuc: i for i, nuc in enumerate(nucs)}
    num_classes = len(mapping)

    # Convert sequence to indices
    indices = np.array([mapping[base] for base in sequence])

    # Create one-hot encoding
    one_hot = np.zeros((len(sequence), num_classes), dtype=int)
    one_hot[np.arange(len(sequence)), indices] = 1

    return one_hot



from collections.abc import Callable

import torch

from mrna_bench.models.embedding_model import EmbeddingModel


class AIDORNA(EmbeddingModel):
    """Inference wrapper for AIDO.RNA.

    AIDO.RNA is a transformer-based RNA foundation model. It is trained using
    masked language modelling on 42 million non-coding RNA sequences, with
    domain adaptation models available for protein coding sequences.

    Link: https://github.com/genbio-ai/ModelGenerator
    """

    MAX_LENGTH = 1024

    @staticmethod
    def get_model_short_name(model_version: str) -> str:
        """Get shortened name of model version."""
        return model_version.replace("rna_", "").replace("_", "-")

    def __init__(self, model_version: str, device: torch.device):
        """Initialize AIDO.RNA.

        Args:
            model_version: Version of model used. Valid versions: {
                "aido_rna_1b600m",
                "aido_rna_1b600m_cds",
                "aido_rna_650m",
                "aido_rna_650m_cds",
            }
            device: PyTorch device to send model to.
        """
        super().__init__(model_version, device)
        from modelgenerator.tasks import Embed

        model = Embed.from_config({"model.backbone": model_version}).eval()

        self.model = model.to(device)

    def embed_sequence(
        self,
        sequence: str,
        overlap: int = 0,
        agg_fn: Callable = torch.mean
    ) -> torch.Tensor:
        """Embed sequence using AIDO.RNA.

        Args:
            sequence: Sequence to be embedded.
            overlap: Number of tokens overlapping between chunks.
            agg_fn: Function used to aggregate embedding across length dim.

        Returns:
            AIDO.RNA embedding of sequence with shape (1 x H).
        """
        chunks = self.chunk_sequence(sequence, self.MAX_LENGTH - 2, overlap)

        embedding_chunks = []

        for i, chunk in enumerate(chunks):
            batch = self.model.transform({"sequences": [chunk]})

            t_keys = ["special_tokens_mask", "input_ids", "attention_mask"]

            # Strip start and stop tokens from all but first and last chunk
            if i == 0:
                for k in t_keys:
                    batch[k] = batch[k][:, :-1]
            elif i == len(chunks) - 1:
                for k in t_keys:
                    batch[k] = batch[k][:, 1:]
            else:
                for k in t_keys:
                    batch[k] = batch[k][:, 1:-1]

            embedded_chunk = self.model(batch)
            embedding_chunks.append(embedded_chunk)

        embedding = torch.cat(embedding_chunks, dim=1)

        aggregate_embedding = agg_fn(embedding, dim=1)
        return aggregate_embedding

    def embed_sequence_sixtrack(self, sequence, cds, splice):
        """Not supported."""
        raise NotImplementedError("Six track not possible with AIDO.RNA.")



from collections.abc import Callable

from mrna_bench.data_splitter.data_splitter import DataSplitter
from mrna_bench.data_splitter.homology_split import HomologySplitter
from mrna_bench.data_splitter.sklearn_split import SklearnSplitter


SPLIT_CATALOG: dict[str, Callable[..., DataSplitter]] = {
    "default": SklearnSplitter,
    "homology": HomologySplitter
}



import os
import requests
from tqdm import tqdm
import yaml


def download_file(
    url: str,
    download_dir: str,
    force_redownload: bool = False
) -> tuple[str, bool]:
    """Download file at the given url.

    Args:
        url: URL of file to be downloaded.
        download_dir: Directory to store downloaded file.
        force_redownload: Forces download even if file already exists.

    Returns:
        Path to downloaded file and whether file was downloaded.
    """
    response = requests.get(url, stream=True)
    response.raise_for_status()
    total_bytes = int(response.headers.get("content-length", 0))

    output_path = download_dir + "/" + os.path.basename(url)

    if os.path.isfile(output_path) and not force_redownload:
        print("File already downloaded.")
        return (output_path, False)

    with open(output_path, "wb") as f:
        with tqdm(total=total_bytes, unit="B", unit_scale=True) as bar:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)
                    bar.update(len(chunk))

    return (output_path, True)


class DataManager:
    """Helper class tracking the directory where benchmark data is stored."""

    def __init__(self):
        """Initialize DataManager."""
        self.curr_dir = os.path.dirname(os.path.abspath(__file__))
        self.config_path = self.curr_dir + "/config.yaml"

    def update_data_path(self, data_dir_path: str):
        """Update path to data storage directory.

        Args:
            data_dir_path: New path to data storage directory.
        """
        data_dir_path = os.path.normpath(data_dir_path)

        config = {"data_path": data_dir_path}

        if not os.path.exists(self.config_path):
            with open(self.config_path, "w") as f:
                yaml.dump(config, f, default_flow_style=False)
        else:
            with open(self.config_path, "r") as f:
                data = yaml.safe_load(f)

            data["data_path"] = data_dir_path

            with open(self.config_path, "w") as f:
                yaml.dump(data, f, default_flow_style=False)

    def get_data_path(self) -> str:
        """Load data_dir_path from config.

        Throws exception if config is not yet initialized.
        """
        if os.path.exists(self.config_path):
            with open(self.config_path) as stream:
                return yaml.safe_load(stream)["data_path"]
        else:
            raise RuntimeError((
                "Data storage path is not set. Please run: "
                "mrna_bench.update_data_path(path_to_store_data)"
            ))


def update_data_path(path_to_data: str):
    """Update path to benchmark data storage directory.

    Args:
        path_to_data: New path to directory where data is stored.
    """
    dm = DataManager()
    dm.update_data_path(path_to_data)


def get_data_path() -> str:
    """Get path where benchmark data is stored.

    Returns:
        Directory where benchmark data is stored.
    """
    dm = DataManager()
    return dm.get_data_path()



from collections.abc import Callable

import torch

from mrna_bench.models import EmbeddingModel


class DNABERT2(EmbeddingModel):
    """Inference wrapper for DNA-BERT2.

    DNABERT2 is a transformer-based DNA foundation model that uses BPE and
    rotary positional encoding among other modern transformer improvements
    to allow for efficient inference. DNABERT2 is pre-trained using MLM
    on multi-species genomic dataset.

    Link: https://github.com/MAGICS-LAB/DNABERT_2
    """

    @staticmethod
    def get_model_short_name(model_version: str) -> str:
        """Get shortened name of model version."""
        return model_version

    def __init__(self, model_version: str, device: torch.device):
        """Initialize DNABERT2 inference wrapper.

        Args:
            model_version: Version of model used; must be "dnabert2".
            device: PyTorch device to send model to.
        """
        super().__init__(model_version, device)

        if model_version != "dnabert2":
            raise ValueError("Only dnabert2 model version available.")

        from transformers import AutoTokenizer, AutoModel

        self.tokenizer = AutoTokenizer.from_pretrained(
            "zhihan1996/DNABERT-2-117M",
            trust_remote_code=True
        )

        self.model = AutoModel.from_pretrained(
            "zhihan1996/DNABERT-2-117M",
            trust_remote_code=True
        ).to(self.device)

    def embed_sequence(
        self,
        sequence: str,
        overlap: int,
        agg_fn: Callable = torch.mean
    ) -> torch.Tensor:
        """Embed sequence using DNABERT2.

        Args:
            sequence: Sequence to embed.
            overlap: Unused.
            agg_fn: Method used to aggregate across sequence dimension.

        Returns:
            DNABERT2 representation of sequence with shape (1 x 768).
        """
        if overlap != 0:
            raise ValueError("DNABERT2 does not chunk sequence.")

        inputs = self.tokenizer(sequence, return_tensors="pt")["input_ids"]
        inputs = inputs.to(self.device)
        hidden_states = self.model(inputs)[0]

        embedding_mean = agg_fn(hidden_states, dim=1)
        return embedding_mean

    def embed_sequence_sixtrack(self, sequence, cds, splice, overlap, agg_fn):
        """Not supported."""
        raise NotImplementedError("Six track not available for DNABERT.")



from collections.abc import Callable

import torch
from transformers import AutoModel, AutoTokenizer

from mrna_bench.models import EmbeddingModel


class HyenaDNA(EmbeddingModel):
    """Inference wrapper for HyenaDNA.

    HyenaDNA is a Hyena-based DNA foundation model trained on the human
    reference genome using an autoregressive scheme at single nucleotide
    resolution. Owing to its state-space backbone, it has an ultra long
    context window.

    Link: https://github.com/HazyResearch/hyena-dna
    """

    @staticmethod
    def get_model_short_name(model_version: str) -> str:
        """Get shortened name of model version."""
        return model_version.replace("-seqlen", "").replace("-hf", "")

    def __init__(self, model_version: str, device: torch.device):
        """Initialize HyenaDNA inference wrapper.

        Support for HyenaDNA 1k models is currently omitted.

        Args:
            model_version: Version of model used. Valid versions are: {
                "hyenadna-large-1m-seqlen-hf",
                "hyenadna-medium-450k-seqlen-hf",
                "hyenadna-medium-160k-seqlen-hf",
                "hyenadna-small-32k-seqlen-hf",
                "hyenadna-tiny-16k-seqlen-d128-hf"
            }
            device: PyTorch device to send model to.
        """
        super().__init__(model_version, device)

        checkpoint = "LongSafari/{}".format(model_version)
        tokenizer = AutoTokenizer.from_pretrained(
            checkpoint,
            trust_remote_code=True
        )

        model = AutoModel.from_pretrained(
            checkpoint,
            torch_dtype=torch.bfloat16,
            device_map="auto",
            trust_remote_code=True
        )

        self.tokenizer = tokenizer
        self.model = model

    def embed_sequence(
        self,
        sequence: str,
        overlap: int,
        agg_fn: Callable = torch.mean
    ) -> torch.Tensor:
        """Embed sequence using HyenaDNA.

        Args:
            sequence: Sequence to embed.
            overlap: Unused.
            agg_fn: Method used to aggregate across sequence dimension.

        Returns:
            HyenaDNA representation of sequence.
        """
        if overlap != 0:
            raise ValueError("HyenaDNA does not chunk sequence.")

        inputs = self.tokenizer(sequence, return_tensors="pt")["input_ids"]
        inputs = inputs.to(self.device)
        hidden_states = self.model(inputs)[0]

        embedding_mean = agg_fn(hidden_states, dim=1)
        return embedding_mean

    def embed_sequence_sixtrack(self, sequence, cds, splice, overlap, agg_fn):
        """Not supported."""
        raise NotImplementedError("Six track not available for HyenaDNA.")



from abc import ABC, abstractmethod
from collections.abc import Callable

import numpy as np
import torch


class EmbeddingModel(ABC):
    """Wrapper class for embedding models used to represent sequences."""

    @staticmethod
    @abstractmethod
    def get_model_short_name(model_version: str) -> str:
        """Retrieve shortened name for model version.

        Should not contain any underscores. Represent spaces with '-'.

        Args:
            model_version: Version of model to fetch short name for.

        Returns:
            Shortened name of model version.
        """
        pass

    def __init__(self, model_version: str, device: torch.device):
        """Initialize EmbeddingModel.

        Args:
            model_version: Version of embedding model to use.
            device: PyTorch device to send embedding model.
        """
        self.model_version = model_version
        self.short_name = self.__class__.get_model_short_name(model_version)
        self.device = device

        self.is_sixtrack = False

        print("Disabling autograd for inference.")
        torch.autograd.set_grad_enabled(False)

    @abstractmethod
    def embed_sequence(
        self,
        sequence: str,
        overlap: int,
        agg_fn: Callable = torch.mean,
    ) -> torch.Tensor:
        """Embed sequence.

        Args:
            sequence: String of nucleotides to embed (uses DNA bases).
            overlap: Number of overlapping nucleotides between chunks.
            agg_fn: Method used to aggregate across sequence dimension.

        Returns:
            Embedded sequence with shape (1 x H).
        """
        pass

    @abstractmethod
    def embed_sequence_sixtrack(
        self,
        sequence: str,
        cds: np.ndarray,
        splice: np.ndarray,
        overlap: int,
        agg_fn: Callable = torch.mean,
    ) -> torch.Tensor:
        """Embed sequence incorporating splice and cds information.

        Args:
            sequence: String of nucleotides to embed (uses DNA bases).
            cds: Binary encoding of first nucleotide of each codon in CDS.
            splice: Binary encoding of splice site locations.
            overlap: Number of overlapping nucleotides between chunks.
            agg_fn: Method used to aggregate across sequence dimension.

        Returns:
            Embedded sequence with shape (1 x H).
        """
        pass

    def chunk_sequence(
        self,
        sequence: str,
        chunk_length: int,
        overlap_size: int = 0
    ) -> list[str]:
        """Split sequence into chunks of specified length with given overlap.

        Args:
            sequence: The input string sequence to be chunked.
            chunk_length: The length of each chunk.
            overlap_size: The number of overlapping characters between chunks.

        Returns:
            A list of string chunks, where each chunk has the specified length.
        """
        if overlap_size >= chunk_length:
            raise ValueError("overlap_size must be less than chunk_length")

        step_size = chunk_length - overlap_size

        chunks = []
        for i in range(0, len(sequence), step_size):
            chunk = sequence[i:i + chunk_length]
            chunks.append(chunk)

        return chunks

    def chunk_tokens(
        self,
        sequence_tokens: list[int],
        chunk_length: int,
        overlap_size: int = 0
    ) -> list[list[int]]:
        """Chunk tokenized sequence into specified length with overlap.

        Args:
            sequence_tokens: The tokenized sequence to be chunked.
            chunk_length: The length of each chunk.
            overlap_size: The number of overlapping tokens between chunks.

        Returns:
            A list of chunked tokens each with specified maximum length.
        """
        if overlap_size >= chunk_length:
            raise ValueError("overlap_size must be less than chunk_length")

        step_size = chunk_length - overlap_size

        chunks = []
        for i in range(0, len(sequence_tokens), step_size):
            chunk = sequence_tokens[i:i + chunk_length]
            chunks.append(chunk)

        return chunks



import pandas as pd

from mrna_bench.datasets.benchmark_dataset import BenchmarkDataset


LNCRNA_URL = "/home/shir2/mRNABench/data/HAP1_essentiality_data.tsv"


class PCGEssentiality(BenchmarkDataset):
    """Protein Coding Gene Essentiality Dataset."""

    def __init__(self, force_redownload: bool = False):
        super().__init__(
            dataset_name="pcg-ess",
            species=["human"],
            raw_data_src_path=LNCRNA_URL,
            force_redownload=force_redownload
        )

    def process_raw_data(self) -> pd.DataFrame:
        data_df = pd.read_csv(self.raw_data_path, delimiter="\t")

        data_df = data_df[data_df["type"] == "pcg"]
        data_df.reset_index(inplace=True, drop=True)
        return data_df



from typing import Type

from .aido import AIDORNA
from .dnabert import DNABERT2
from .hyenadna import HyenaDNA
from .nucleotide_transformer import NucleotideTransformer
from .orthrus import Orthrus
from .rnafm import RNAFM

from .embedding_model import EmbeddingModel


MODEL_CATALOG: dict[str, Type[EmbeddingModel]] = {
    "AIDO.RNA": AIDORNA,
    "DNABERT2": DNABERT2,
    "NucleotideTransformer": NucleotideTransformer,
    "Orthrus": Orthrus,
    "RNA-FM": RNAFM,
    "HyenaDNA": HyenaDNA
}



def get_output_filename(
    output_dir: str,
    model_version: str,
    dataset_name: str,
    sequence_chunk_overlap: int,
    d_chunk_ind: int = 0,
    d_num_chunks: int = 0
) -> str:
    """Get standardized embedding file name.

    NOTE: Model and dataset names should not have underscores.

    Args:
        output_dir: Directory to store embeddings.
        model_version: Name of embedding model version.
        dataset_name: Dataset which is embedded.
        sequence_chunk_overlap: Number of tokens overlapped in sequence chunks.
        d_chunk_ind: Index of current dataset chunk.
        d_num_chunks: Maximum number of dataset chunks.
    """
    out_path = "{}/{}_{}_o{}".format(
        output_dir,
        dataset_name,
        model_version,
        sequence_chunk_overlap
    )

    if d_num_chunks != 0:
        out_path += "_{}-{}".format(d_chunk_ind, d_num_chunks)

    return out_path



from collections.abc import Callable

from .benchmark_dataset import BenchmarkDataset
from .go_mol_func import GOMolecularFunction
from .pcg_essentiality import PCGEssentiality
from .lncrna_essentiality import LNCRNAEssentiality
from .rna_hl_human import RNAHalfLifeHuman
from .rna_hl_mouse import RNAHalfLifeMouse
from .prot_loc import ProteinLocalization
from .mrl_sugimoto import MRLSugimoto

DATASET_CATALOG: dict[str, Callable[..., BenchmarkDataset]] = {
    "go-mf": GOMolecularFunction,
    "pcg-ess": PCGEssentiality,
    "lncrna-ess": LNCRNAEssentiality,
    "rnahl-human": RNAHalfLifeHuman,
    "rnahl-mouse": RNAHalfLifeMouse,
    "prot-loc": ProteinLocalization,
    "mrl-sugimoto": MRLSugimoto,
}



from collections.abc import Callable

import torch

from mrna_bench.models import EmbeddingModel


class NucleotideTransformer(EmbeddingModel):
    """Inference wrapper for NucleotideTransformer.

    NucleotideTransformer is a transformer based DNA foundation model
    pre-trained using MLM on a variety of pre-training datasets ranging from
    the 1000 (human) genomes project to a multi-species dataset, across
    a wide range of parameters. Input is tokenized to 6-mers where possible,
    with a maximum tokenized sequence length of 1000.

    Link: https://github.com/instadeepai/nucleotide-transformer
    """

    @staticmethod
    def get_model_short_name(model_version: str) -> str:
        """Get shortened name of model version."""
        return "nt-" + model_version

    def __init__(self, model_version: str, device: torch.device):
        """Initialize NucleotideTransformer inference wrapper.

        Args:
            model_version: Version of model to load. Valid versions are: {
                "2.5b-multi-species",
                "2.5b-1000g",
                "500m-human-ref",
                "500m-1000g",
                "v2-50m-multi-species",
                "v2-100m-multi-species",
                "v2-250m-multi-species",
                "v2-500m-multi-species"
            }
            device: PyTorch device to send model to.
        """
        super().__init__(model_version, device)

        from transformers import AutoTokenizer, AutoModelForMaskedLM

        self.tokenizer = AutoTokenizer.from_pretrained(
            "InstaDeepAI/nucleotide-transformer-{}".format(model_version),
            trust_remote_code=True
        )

        self.model = AutoModelForMaskedLM.from_pretrained(
            "InstaDeepAI/nucleotide-transformer-{}".format(model_version),
            trust_remote_code=True
        ).to(self.device)

        self.max_length = self.tokenizer.model_max_length

    def embed_sequence(
        self,
        sequence: str,
        overlap: int,
        agg_fn: Callable = torch.mean
    ) -> torch.Tensor:
        """Embed sequence using NucleotideTransformer.

        Args:
            sequence: Sequence to be embedded.
            overlap: Number of tokens overlapping between chunks.
            agg_fn: Function used to aggregate embedding across length dim.

        Returns:
            NT embedding of sequence with shape (1 x H).
        """
        tokenized = self.tokenizer.encode_plus(sequence, verbose=False)

        chunks = self.chunk_tokens(
            tokenized["input_ids"],
            self.max_length,
            overlap
        )

        embedding_chunks = []

        for _, chunk in enumerate(chunks):
            chunk_tt = torch.Tensor(chunk).unsqueeze(0).to(self.device).long()
            attention_mask = torch.ones_like(chunk_tt).to(self.device).long()

            torch_outs = self.model(
                chunk_tt,
                attention_mask=attention_mask,
                encoder_attention_mask=attention_mask,
                output_hidden_states=True
            )

            model_out = torch_outs["hidden_states"][-1]
            embedding_chunks.append(model_out)

        embedding = torch.cat(embedding_chunks, dim=1)

        aggregate_embedding = agg_fn(embedding, dim=1)
        return aggregate_embedding

    def embed_sequence_sixtrack(self, sequence, cds, splice, overlap, agg_fn):
        """Not supported."""
        raise NotImplementedError("Six track not available for NT.")



from collections.abc import Callable

import numpy as np
import torch

from transformers import AutoModel

from mrna_bench.models import EmbeddingModel


class Orthrus(EmbeddingModel):
    """Inference wrapper for Orthrus.

    Orthrus is a RNA foundation model trained using a Mamba backbone. It uses
    a contrastive learning pre-training objective that maximizes similarity
    between RNA splice isoforms and orthologous transcripts. Input length is
    unconstrained due to use of Mamba.

    Link: https://github.com/bowang-lab/Orthrus
    """

    @staticmethod
    def get_model_short_name(model_version: str) -> str:
        """Get shortened name of model version."""
        return model_version.replace("-track", "")

    def __init__(self, model_version: str, device: torch.device):
        """Initialize Orthrus model.

        Args:
            model_version: Version of Orthrus to load. Valid values are: {
                "orthrus-base-4-track",
                "orthrus-large-6-track"
            }
            device: PyTorch device to send model to.
        """
        super().__init__(model_version, device)

        model_hf_path = "quietflamingo/{}".format(model_version)
        model = AutoModel.from_pretrained(
            model_hf_path,
            trust_remote_code=True
        )

        self.is_sixtrack = model_version == "orthrus-large-6-track"
        self.model = model.to(device)

    def embed_sequence(
        self,
        sequence: str,
        overlap: int = 0,
        agg_fn: Callable | None = None
    ) -> torch.Tensor:
        """Embed sequence using four track Orthrus.

        Args:
            sequence: Sequence to embed.
            overlap: Unused.
            agg_fn: Currently unused.

        Returns:
            Orthrus representation of sequence.
        """
        if overlap != 0:
            raise ValueError("Orthrus does not chunk sequence.")

        if agg_fn is not None:
            raise NotImplementedError(
                "Inference currently does not support alternative aggregation."
            )

        ohe_sequence = self.model.seq_to_oh(sequence).to(self.device)
        model_input_tt = ohe_sequence.unsqueeze(0)

        lengths = torch.Tensor([model_input_tt.shape[1]]).to(self.device)

        embedding = self.model.representation(
            model_input_tt,
            lengths,
            channel_last=True
        )

        return embedding

    def embed_sequence_sixtrack(
        self,
        sequence: str,
        cds: np.ndarray,
        splice: np.ndarray,
        overlap: int = 0,
        agg_fn: Callable | None = None,
    ) -> torch.Tensor:
        """Embed sequence using six track Orthrus.

        Expects binary encoded tracks denoting the beginning of each codon
        in the CDS and the 5' ends of each splice site.

        Args:
            sequence: Sequence to embed.
            cds: CDS track for sequence to embed.
            splice: Splice site track for sequence to embed.
            overlap: Unused.
            agg_fn: Currently unused.

        Returns:
            Orthrus representation of sequence.
        """
        if overlap != 0:
            raise ValueError("Orthrus does not chunk sequence.")

        if agg_fn is not None:
            raise NotImplementedError(
                "Inference currently does not support alternative aggregation."
            )

        ohe_sequence = self.model.seq_to_oh(sequence).numpy()

        model_input = np.hstack((
            ohe_sequence,
            cds.reshape(-1, 1),
            splice.reshape(-1, 1)
        ))

        model_input_tt = torch.Tensor(model_input).to(self.device)
        model_input_tt = model_input_tt.unsqueeze(0)

        lengths = torch.Tensor([model_input_tt.shape[1]]).to(self.device)

        embedding = self.model.representation(
            model_input_tt,
            lengths,
            channel_last=True
        )

        return embedding



import numpy as np
import pandas as pd

from mrna_bench.datasets.benchmark_dataset import BenchmarkDataset
from mrna_bench.datasets.dataset_utils import ohe_to_str


RNAHLM_URL = "https://zenodo.org/records/14708163/files/rna_hl_mouse.npz"


class RNAHalfLifeMouse(BenchmarkDataset):
    """RNA Halflife in Mouse Dataset."""

    def __init__(self, force_redownload: bool = False):
        """Initialize RNAHalfLifeMouse dataset.

        Args:
            force_redownload: Force raw data download even if pre-existing.
        """
        super().__init__(
            dataset_name="rnahl-mouse",
            species=["mouse"],
            raw_data_src_url=RNAHLM_URL,
            force_redownload=force_redownload
        )

    def process_raw_data(self) -> pd.DataFrame:
        """Process raw data into Pandas dataframe.

        Returns:
            Pandas dataframe of processed sequences.
        """
        data = np.load(self.raw_data_path)

        X = data["X"]

        seq_str = ohe_to_str(X[:, :, :4])
        lens = [len(s) for s in seq_str]
        cds = [X[i, :lens[i], 4] for i in range(len(X))]
        splice = [X[i, :lens[i], 5] for i in range(len(X))]

        df = pd.DataFrame({
            "sequence": seq_str,
            "target": [y for y in data["y"]],
            "gene": data["genes"],
            "transcript_length": lens,
            "cds": cds,
            "splice": splice
        })

        return df



from abc import ABC, abstractmethod
from pathlib import Path
import shutil
import os

import pandas as pd

from mrna_bench.utils import download_file, get_data_path


class BenchmarkDataset(ABC):
    """Abstract class for benchmarking datasets.

    Sequences are internally represented as strings. This is less storage
    efficient, but easier to handle as most parts of the pipeline like to
    use raw text.

    TODO: The inheritance pattern here is kind of wonky. Probably best to
    make it so each task has a BenchmarkDataset, but inherits from something
    else.
    """

    def __init__(
        self,
        dataset_name: str,
        species: list[str] = ["human"],
        raw_data_src_url: str | None = None,
        force_redownload: bool = False,
        raw_data_src_path: str | None = None,
    ):
        """Initialize BenchmarkDataset.

        Args:
            dataset_name: Name of the benchmark dataset. Should have no
                spaces, use '-' instead.
            species: Species dataset is collected from.
            raw_data_src_url: URL where raw data can be downloaded.
            force_redownload: Forces raw data redownload.
            raw_data_src_path: Path where raw data is located.
        """
        if raw_data_src_url is None and raw_data_src_path is None:
            raise ValueError("At least one data source must be defined.")
        elif raw_data_src_path is not None and raw_data_src_url is not None:
            raise ValueError("Only one data source must be defined.")

        self.dataset_name = dataset_name

        self.raw_data_src_url = raw_data_src_url
        self.raw_data_src_path = raw_data_src_path

        self.species = species

        self.force_redownload = force_redownload

        self.data_storage_path = get_data_path()
        self.init_folders()

        if force_redownload or not self.load_processed_df():
            if self.raw_data_src_url is None:
                self.collect_raw_data()
            else:
                self.download_raw_data()

            # TODO: This bugs out due to parallelism I think.
            self.data_df = self.process_raw_data()
            self.save_processed_df(self.data_df)

    def init_folders(self):
        """Initialize folders for storing raw data.

        Creates a structure with:

        - data_path
        |    - dataset_name
        |    |    - raw_data
        |    |    - embeddings
        """
        ds_path = Path(self.data_storage_path) / self.dataset_name
        ds_path.mkdir(exist_ok=True)

        raw_data_dir = Path(ds_path) / "raw_data"
        raw_data_dir.mkdir(exist_ok=True)

        emb_dir = Path(ds_path) / "embeddings"
        emb_dir.mkdir(exist_ok=True)

        self.dataset_path = str(ds_path)
        self.raw_data_dir = str(raw_data_dir)
        self.embedding_dir = str(emb_dir)

    def download_raw_data(self):
        """Download the raw data from given web source."""
        raw_data_path, _ = download_file(
            self.raw_data_src_url,
            self.raw_data_dir,
            self.force_redownload
        )
        self.raw_data_path = raw_data_path

    def collect_raw_data(self):
        """Collect the raw data from given local path."""
        raw_file_name = Path(self.raw_data_src_path).name
        raw_data_path = self.raw_data_dir + "/" + raw_file_name

        if not os.path.exists(raw_data_path):
            shutil.copy(self.raw_data_src_path, raw_data_path)

        self.raw_data_path = raw_data_path

    def save_processed_df(self, df: pd.DataFrame):
        """Save dataframe to data storage path.

        Args:
            df: Processed dataframe to save.
        """
        df.to_pickle(self.dataset_path + "/data_df.pkl")

    def load_processed_df(self) -> bool:
        """Load processed dataframe from data storage path.

        Returns:
            Whether dataframe was successfully loaded to class property.
        """
        try:
            self.data_df = pd.read_pickle(self.dataset_path + "/data_df.pkl")
        except FileNotFoundError:
            print("Processed data frame not found.")
            return False
        return True

    @abstractmethod
    def process_raw_data(self) -> pd.DataFrame:
        """Abstract method to process the dataset for the task."""
        pass



from pathlib import Path
import shutil
import zipfile

import pandas as pd
import numpy as np

from mrna_bench import get_data_path
from mrna_bench.data_splitter.data_splitter import DataSplitter
from mrna_bench.utils import download_file


def train_test_split_homologous(
    genes: list[str],
    homology_df: pd.DataFrame,
    test_size: float = 0.3,
    random_state: int | None = None
) -> tuple[list[int], list[int]]:
    """Split genes into two sets with homologous genes in the same set.

    Args:
        genes: List of gene names.
        homology_df: DataFrame with columns "gene_name" and "gene_group".
        test_size: Fraction of data allocated to test split.
        random_state: Defaults to None.

    Returns:
        Dictionary with keys "train_indices" and "test_indices" containing the
        indices of the genes in the train and test sets respectively.
    """
    # Map genes to their respective homology groups
    homo_group_map = homology_df.set_index("gene_name")["gene_group"].to_dict()

    gene_groups = np.array([homo_group_map.get(gene, None) for gene in genes])

    group_to_index: dict[int, list[int]] = {}

    # Populate the group_to_index dictionary
    for i, group in enumerate(gene_groups):
        if group is not None:
            group_to_index.setdefault(group, []).append(i)

    gene_index = np.arange(len(genes))

    np.random.seed(random_state)
    np.random.shuffle(gene_index)

    len_of_train = int(len(gene_index) * (1 - test_size))

    train_indices: list[int] = []
    test_indices: list[int] = []

    seen_groups = set()

    for index in gene_index:
        group = gene_groups[index]
        if group is not None and group in seen_groups:
            continue

        seen_groups.add(group)

        if len(train_indices) < len_of_train:
            train_indices.extend(group_to_index.get(group, [index]))
        else:
            test_indices.extend(group_to_index.get(group, [index]))

    train_indices = [int(ind) for ind in train_indices]
    test_indices = [int(ind) for ind in test_indices]

    return train_indices, test_indices


class HomologySplitter(DataSplitter):
    """Homology-based data splitter.

    Uses an external homology mapping file to construct train / test splits.
    Genes which are homologous are kept within the same 'side' of the data
    split to reduce data leakage.
    """

    HOMO_URL = (
        "https://zenodo.org/records/13910050/files/"
        "homology_maps_homologene.zip"
    )

    def __init__(
        self,
        species: list[str],
        homology_map_path: str | None = None,
        force_redownload: bool = False
    ):
        """Initialize HomologySplitter.

        Homology splitting requires a dataframe mapping genes to a group
        of homologous genes. This dataframe must have columns "gene_name" and
        "gene_group". If homology_map_path is not specified, relationships
        processed from homologene will be downloaded.

        Homology map files should be named as: {species}_homology_map.csv

        Args:
            species: List of species that genes are from.
            homology_map_path: Path to homology maps.
            force_redownload: Forces redownload of homology maps.
        """
        super().__init__()

        if homology_map_path is None:
            data_storage_path = get_data_path()
            self.homology_map_path = data_storage_path + "/homology_maps"

            if not Path(self.homology_map_path).exists():
                Path(self.homology_map_path).mkdir(exist_ok=True)
                force_redownload = True
        else:
            self.homology_map_path = homology_map_path

        if force_redownload:
            out = download_file(self.HOMO_URL, self.homology_map_path, True)[0]

            with zipfile.ZipFile(out, "r") as zip_ref:
                for file in zip_ref.namelist():
                    if not file.endswith("/"):
                        source = zip_ref.open(file)
                        path = Path(self.homology_map_path) / Path(file).name

                        with open(path, "wb") as target:
                            shutil.copyfileobj(source, target)

            Path(out).unlink()

        self.species = species
        self.homology_df = self.get_homology_df()

    def get_homology_df(self) -> pd.DataFrame:
        """Get homology dataframe for specified species.

        If multiple species exists, naively merges dataframes, merging homology
        groups if gene name overlap exists.

        Returns:
            Dataframe with gene names mapped to homology groups.
        """
        species_maps = []
        for s in self.species:
            homology_dir = Path(self.homology_map_path)
            homology_map = homology_dir / "{}_homology_map.csv".format(s)
            if not homology_map.exists():
                print("Homology map for species {} is missing.".format(s))
            else:
                species_maps.append(pd.read_csv(homology_map))

        max_groups = [max(s_df["gene_group"]) for s_df in species_maps]
        cusum_max_groups = np.cumsum(max_groups)

        for i, s_df in enumerate(species_maps[1:]):
            s_df["gene_group"] += cusum_max_groups[i]

        out_df = pd.concat(species_maps, axis=0, ignore_index=True)

        # Merge identical gene names
        out_df.groupby("gene_name", as_index=False).min()
        out_df.reset_index(drop=True, inplace=True)

        return out_df

    def split_df(
        self,
        df: pd.DataFrame,
        test_size: float,
        random_seed: int
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Split dataframe rows into train and test df using homology split.

        Args:
            df: Dataframe to split.
            test_size: Fraction of dataset to assign to test split.
            random_seed: Random seed used for sampling rows during splitting.

        Returns:
            Dataframe containing train data and test data.
        """
        if "gene" not in df.columns:
            raise ValueError("Gene must be column for homology split.")

        genes = df["gene"]

        train_indices, test_indices = train_test_split_homologous(
            genes,
            self.homology_df,
            test_size,
            random_seed
        )

        assert len(set(train_indices).intersection(set(test_indices))) == 0

        train_df = df.iloc[train_indices]
        test_df = df.iloc[test_indices]

        return train_df, test_df



from abc import ABC, abstractmethod

import pandas as pd


class DataSplitter(ABC):
    """Generates reproducible train test splits."""

    @abstractmethod
    def split_df(
        self,
        df: pd.DataFrame,
        test_size: float,
        random_seed: int
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Split dataframe into train and test dataframes.

        Args:
            df: Dataframe to split.
            test_size: Fraction of dataset to assign to test split.
            random_seed: Random seed used during split sampling.

        Returns:
            Train and test dataframes.
        """
        pass

    def get_all_splits_df(
        self,
        df: pd.DataFrame,
        split_ratios: tuple[float, float, float],
        random_seed: int,
    ) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """Get train, validation, test splits.

        Args:
            df: Dataframe to split.
            split_ratios: Ratio of training, validation, test split size.
            random_seed: Random seed used to generate splits.

        Returns:
            Dataframe containing training, validation, test splits.
        """
        if sum(split_ratios) != 1:
            raise ValueError("Split ratios must sum to 1.")

        tv_split_size = split_ratios[1] + split_ratios[2]
        test_split_size = split_ratios[2] / tv_split_size

        train_df, tv_df = self.split_df(df, tv_split_size, random_seed)
        val_df, test_df = self.split_df(tv_df, test_split_size, random_seed)

        return train_df, val_df, test_df



from collections.abc import Callable

import numpy as np
import torch

from mrna_bench.models.embedding_model import EmbeddingModel


class RNAFM(EmbeddingModel):
    """Inference Wrapper for RNA-FM.

    RNA-FM is a transformer based RNA foundation model pre-trained using MLM on
    23 million ncRNA sequences. The primary competency for RNA-FM is ncRNA
    property and structural prediction.

    mRNA-FM is a related model that is instead pre-trained on coding sequences.
    It can only accept CDS regions (input must be multiple of 3).

    Link: https://github.com/ml4bio/RNA-FM/
    """

    MAX_LENGTH = 1024

    @staticmethod
    def get_model_short_name(model_version: str) -> str:
        """Get shortened name of model version."""
        return model_version

    def __init__(self, model_version: str, device: torch.device):
        """Initialize RNA-FM Model.

        Args:
            model_version: Version of RNA-FM to use. Valid versions are:
                {"rna-fm", "mrna-fm"}.
            device: PyTorch device used by model inference.
        """
        super().__init__(model_version, device)

        import fm

        if model_version == "rna-fm":
            model, alphabet = fm.pretrained.rna_fm_t12()
            self.is_sixtrack = False
        elif model_version == "mrna-fm":
            model, alphabet = fm.pretrained.mrna_fm_t12()
            self.is_sixtrack = True
        else:
            raise ValueError("Unknown model version.")

        self.model = model.to(device).eval()
        self.batch_converter = alphabet.get_batch_converter()

    def embed_sequence(
        self,
        sequence: str,
        overlap: int = 0,
        agg_fn: Callable = torch.mean
    ) -> torch.Tensor:
        """Embed sequence using RNA-FM.

        Due to RNA-FM's max context being shorter than most mRNAs, chunking is
        used. Here, sequence is chunked, and start / end tokens are stripped
        from the middle sequences. Representations are then averaged across
        the sequence length dimension.

        Args:
            sequence: Sequence to embed.
            overlap: Number of overlapping nucleotides between chunks.
            agg_fn: Method used to aggregate across sequence dimension.

        Returns:
            RNA-FM representation of sequence with shape (1 x 640).
        """
        sequence = sequence.replace("U", "T")
        chunks = self.chunk_sequence(sequence, self.MAX_LENGTH - 2, overlap)

        embedding_chunks = []

        for i, chunk in enumerate(chunks):
            _, _, tokens = self.batch_converter([("", chunk)])

            if i == 0:
                tokens = tokens[:, :-1]
            elif i == len(chunks) - 1:
                tokens = tokens[:, 1:]
            else:
                tokens = tokens[:, 1:-1]

            model_output = self.model(tokens.to(self.device), repr_layers=[12])
            embedded_chunk = model_output["representations"][12]

            embedding_chunks.append(embedded_chunk)

        embedding = torch.cat(embedding_chunks, dim=1)

        aggregate_embedding = agg_fn(embedding, dim=1)
        return aggregate_embedding

    def embed_sequence_sixtrack(
        self,
        sequence: str,
        cds: np.ndarray,
        splice: np.ndarray,
        overlap: int,
        agg_fn: Callable = torch.mean,
    ) -> torch.Tensor:
        """Embed sequence using mRNA-FM.

        Since mRNA-FM only accepts CDS, uses CDS track to extract CDS sequence
        and generate representation from it. CDS sequence must be a multiple
        of three, complicating chunking with overlap. It is disabled for now.

        Args:
            sequence: Sequence to embed.
            cds: Binary encoding of first nucleotide of each codon in CDS.
            splice: Binary encoding of splice site locations.
            overlap: Number of overlapping nucleotides between chunks.
            agg_fn: Method used to aggregate across sequence dimension.

        Returns:
            mRNA-FM representation of CDS of sequence with shape (1 x H).
        """
        if overlap != 0:
            raise ValueError("mRNA-FM wrapper does not support overlap.")

        _ = splice, overlap  # unused

        sequence = sequence.replace("U", "T")

        first_one_index = np.argmax(cds == 1)
        last_one_index = (len(cds) - 1 - np.argmax(np.flip(cds) == 1)) + 2

        cds_seq = sequence[first_one_index:last_one_index + 1]

        if len(cds_seq) % 3 != 0:
            raise ValueError("Length of CDS is not a multiple of 3.")

        chunks = self.chunk_sequence(cds_seq, 1022 * 3)

        embedding_chunks = []

        for i, chunk in enumerate(chunks):
            _, _, tokens = self.batch_converter([("", chunk)])

            if i == 0:
                tokens = tokens[:, :-1]
            elif i == len(chunks) - 1:
                tokens = tokens[:, 1:]
            else:
                tokens = tokens[:, 1:-1]

            model_output = self.model(tokens.to(self.device), repr_layers=[12])
            embedded_chunk = model_output["representations"][12]

            embedding_chunks.append(embedded_chunk)

        embedding = torch.cat(embedding_chunks, dim=1)

        aggregate_embedding = agg_fn(embedding, dim=1)
        return aggregate_embedding



from pathlib import Path

import numpy as np
import pandas as pd
from tqdm import tqdm

import torch

from mrna_bench.models import EmbeddingModel
from mrna_bench.datasets import BenchmarkDataset
from mrna_bench.embedder.embedder_utils import get_output_filename


class DatasetEmbedder:
    """Embeds sequences associated with dataset using specified embedder.

    This class is built to split the sequences in a dataset into chunks of
    sequences which can then be processed in parallel. This is denoted d_chunk,
    while s_chunk denotes the sequence chunking that occur within each model
    to handle sequences that exceed model maximum length.
    """

    def __init__(
        self,
        model: EmbeddingModel,
        dataset: BenchmarkDataset,
        s_chunk_overlap: int = 0,
        d_chunk_ind: int = 0,
        d_num_chunks: int = 0,
    ):
        """Initialize DatasetEmbedder.

        Args:
            model: Model used to embed sequences.
            dataset: Dataset to embed.
            s_chunk_overlap: Number of overlapping tokens between chunks in
                individual sequences when using chunking to handle input
                exceeding maximum model length.
            d_chunk_ind: Current dataset chunk to be processed.
            d_num_chunks: Total number of chunks to divide dataset into.
        """
        self.model = model
        self.dataset = dataset
        self.data_df = dataset.data_df
        self.s_chunk_overlap = s_chunk_overlap

        self.d_chunk_ind = d_chunk_ind
        self.d_num_chunks = d_num_chunks

        if self.d_num_chunks == 0:
            self.d_chunk_size = len(self.data_df)
        else:
            self.d_chunk_size = (len(self.data_df) // self.d_num_chunks) + 1

    def get_dataset_chunk(self) -> pd.DataFrame:
        """Retrieve current dataset chunk to be embedded.

        Returns:
            Current dataset chunk to be embedded.
        """
        if self.d_num_chunks == 0:
            return self.data_df

        s = self.d_chunk_size * self.d_chunk_ind
        e = s + self.d_chunk_size

        chunk_df = self.data_df.iloc[s:e]
        return chunk_df

    def embed_dataset(self) -> torch.Tensor:
        """Compute embeddings for current dataset chunk.

        Returns:
            Embeddings for current dataset chunk in original order.
        """
        dataset_chunk = self.get_dataset_chunk()

        dataset_embeddings = []
        for _, row in tqdm(dataset_chunk.iterrows(), total=len(dataset_chunk)):
            if self.model.is_sixtrack:
                embedding = self.model.embed_sequence_sixtrack(
                    row["sequence"],
                    row["cds"].astype(np.int32),
                    row["splice"].astype(np.int32),
                    self.s_chunk_overlap,
                )
            else:
                embedding = self.model.embed_sequence(
                    row["sequence"],
                    self.s_chunk_overlap,
                )
            dataset_embeddings.append(embedding)

        embeddings = torch.cat(dataset_embeddings, dim=0)
        return embeddings

    def persist_embeddings(self, embeddings: torch.Tensor):
        """Persist embeddings at global data storage location.

        Args:
            embedding: Embedding to persist.
        """
        out_path = get_output_filename(
            self.dataset.embedding_dir,
            self.model.short_name,
            self.dataset.dataset_name,
            self.s_chunk_overlap,
            self.d_chunk_ind,
            self.d_num_chunks
        )

        np_embeddings = embeddings.float().detach().cpu().numpy()
        np.savez_compressed(out_path, embedding=np_embeddings)

    def merge_embeddings(self):
        """Merge persisted processed dataset chunks into single file.

        Process will only complete if all chunks are finished processing.
        """
        all_chunks = list(range(self.d_num_chunks))
        processed_files_paths = []
        processed_chunk_inds = []

        # Check that all chunks are processed
        for file in Path(self.dataset.embedding_dir).iterdir():
            if not file.is_file():
                continue

            file_name = file.stem
            file_name_arr = file_name.split("_")

            if file_name_arr[0] != self.dataset.dataset_name:
                continue
            if file_name_arr[1] != self.model.short_name:
                continue
            if int(file_name_arr[2][1:]) != self.s_chunk_overlap:
                continue

            chunk_coords = file_name_arr[3].split("-")
            if int(chunk_coords[-1]) != self.d_num_chunks:
                continue

            processed_chunk_inds.append(int(chunk_coords[0]))
            processed_files_paths.append(file)

        if len(set(all_chunks) - set(processed_chunk_inds)) > 0:
            return

        print("All embedding chunks computed. Merging.")

        processed_files_paths = sorted(
            processed_files_paths,
            key=lambda x: int(Path(x).stem.split("_")[-1].split("-")[0])
        )

        embeddings = []
        for file_path in processed_files_paths:
            embedding_chunk = np.load(file_path)["embedding"]
            embeddings.append(embedding_chunk)

        all_embeddings = np.concatenate(embeddings, axis=0)

        out_fn = get_output_filename(
            self.dataset.embedding_dir,
            self.model.short_name,
            self.dataset.dataset_name,
            self.s_chunk_overlap
        )

        np.savez_compressed(out_fn, embedding=all_embeddings)

        for file in processed_files_paths:
            Path(file).unlink()



from typing import TYPE_CHECKING, Type

from mrna_bench.datasets import BenchmarkDataset, DATASET_CATALOG

if TYPE_CHECKING:
    import torch
    from mrna_bench.models import EmbeddingModel


def load_model(
    model_name: str,
    model_version: str,
    device: "torch.device"
) -> "EmbeddingModel":
    """Load Embedding Model.

    Args:
        model_name: Name of model class.
        model_version: Specific model version to load.
        device: PyTorch device to load model to.

    Returns:
        Initialized EmbeddingModel.
    """
    try:
        from mrna_bench.models import EmbeddingModel, MODEL_CATALOG

        model_class: Type[EmbeddingModel] = MODEL_CATALOG[model_name]
        model = model_class(model_version, device)
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "PyTorch not installed. Model benchmarking unavailable."
        )

    return model


def load_dataset(
    dataset_name: str,
    force_redownload: bool = False
) -> BenchmarkDataset:
    """Load Benchmark Dataset.

    Args:
        dataset_name: Name of the dataset.
        force_redownload: Forces data file redownload.
    """
    return DATASET_CATALOG[dataset_name](force_redownload)



from mrna_bench.loader.loader import load_model, load_dataset
from mrna_bench.utils import update_data_path, get_data_path

__all__ = [
    "load_model",
    "load_dataset",
    "update_data_path",
    "get_data_path"
]



import numpy as np
from scipy.stats import pearsonr

from sklearn.base import RegressorMixin, ClassifierMixin
from sklearn.linear_model import RidgeCV, LinearRegression, LogisticRegression
from sklearn.multioutput import MultiOutputClassifier
from sklearn.metrics import average_precision_score, roc_auc_score

from mrna_bench import load_dataset
from mrna_bench.data_splitter.data_splitter import DataSplitter
from mrna_bench.data_splitter.split_catalog import SPLIT_CATALOG
from mrna_bench.datasets import BenchmarkDataset
from mrna_bench.models import EmbeddingModel, MODEL_CATALOG
from mrna_bench.embedder import get_output_filename


class LinearProbe:
    """Linear Probe Evaluation Module.

    Performs linear probing on embeddings for sequences from a benchmark
    dataset generated by an embedding model. The linear probing task can
    be selected at initialization on a target column from the benchmarking
    dataset dataframe.

    The train and test splits for linear probing can be recomputed from random
    seeds. This module supports generating splits based on sequence gene name
    homology from homologene.
    """

    linear_models = {
        "regression": RidgeCV(alphas=[1e-3, 1e-2, 1e-1, 1, 10]),
        "reg_lin": LinearRegression(),
        "reg_ridge": RidgeCV(alphas=[1e-3, 1e-2, 1e-1, 1, 10]),
        "classification": LogisticRegression(max_iter=5000),
        "multilabel": MultiOutputClassifier(
            LogisticRegression(max_iter=5000)
        )
    }

    @staticmethod
    def load_persisted_embeddings(
        embedding_dir: str,
        model_short_name: str,
        dataset_name: str,
        seq_overlap: int = 0
    ) -> np.ndarray:
        """Load pre-computed embeddings for dataset from persisted location.

        Args:
            embedding_dir: Directory where embedding is stored.
            model_short_name: Shortened name of embedding model version.
            dataset_name: Name of dataset which was embedded.
            seq_overlap: Sequence chunking overlap used during embedding.

        Returns:
            Embeddings for dataset computed using embedding model.
        """
        embeddings_fn = get_output_filename(
            embedding_dir,
            model_short_name,
            dataset_name,
            seq_overlap,
        ) + ".npz"

        embeddings = np.load(embeddings_fn)["embedding"]
        return embeddings

    @classmethod
    def init_from_instance(
        cls,
        model: EmbeddingModel,
        dataset: BenchmarkDataset,
        task: str,
        target_col: str,
        seq_chunk_overlap: int = 0,
        split_type: str = "homology",
        split_ratios: tuple[float, float, float] = (0.7, 0.15, 0.15),
        eval_all_splits: bool = False
    ) -> "LinearProbe":
        """Initialize LinearProbe from instantiated dataset and model.

        Args:
            model: EmbeddingModel that generated embeddings to be probed.
            dataset: BenchmarkDataset containing embedded sequences.
            task: Linear probing task. Valid tasks: {"regression", "reg_lin",
                "reg_ridge", "classification", "multilabel"}.
            target_col: Column from dataframe to use as labels.
            seq_chunk_overlap: Sequence chunk overlap used during embedding.
            split_type: Method used for data split generation. Valid types:
                {"default", "homology"}.
            split_ratios: Ratio of data split sizes as a fraction of dataset.
            eval_all_splits: Evaluate metrics on all splits. Only evaluates
                validation split otherwise.

        Returns:
            Initialized LinearProbe.
        """
        embeddings = cls.load_persisted_embeddings(
            dataset.embedding_dir,
            model.short_name,
            dataset.dataset_name,
            seq_chunk_overlap
        )
        return cls(
            dataset,
            embeddings,
            task,
            target_col,
            split_type,
            split_ratios,
            eval_all_splits
        )

    @classmethod
    def init_from_name(
        cls,
        model_name: str,
        model_version: str,
        dataset_name: str,
        task: str,
        target_col: str,
        seq_chunk_overlap: int = 0,
        split_type: str = "homology",
        split_ratios: tuple[float, float, float] = (0.7, 0.15, 0.15),
        eval_all_splits: bool = False
    ) -> "LinearProbe":
        """Initialize LinearProbe using model and dataset names.

        Args:
            model_name: Name of model to be probed.
            model_version: Version of model to be probed.
            dataset_name: Name of dataset containing embedded sequences.
            task: Linear probing task. Valid tasks: {"regression", "reg_lin",
                "reg_ridge", "classification", "multilabel"}.
            target_col: Column from dataframe to use as labels.
            seq_chunk_overlap: Sequence chunk overlap used during embedding.
            split_type: Method used for data split generation. Valid types:
                {"default", "homology"}.
            split_ratios: Ratio of data split sizes as a fraction of dataset.
            eval_all_splits: Evaluate metrics on all splits. Only evaluates
                validation split otherwise.

        Returns:
            Initialized LinearProbe.
        """
        model_class = MODEL_CATALOG[model_name]
        model_short_name = model_class.get_model_short_name(model_version)

        dataset = load_dataset(dataset_name)

        embeddings = cls.load_persisted_embeddings(
            dataset.embedding_dir,
            model_short_name,
            dataset_name,
            seq_chunk_overlap
        )

        return cls(
            dataset,
            embeddings,
            task,
            target_col,
            split_type,
            split_ratios,
            eval_all_splits
        )

    def __init__(
        self,
        dataset: BenchmarkDataset,
        embeddings: np.ndarray,
        task: str,
        target_col: str,
        split_type: str = "homology",
        split_ratios: tuple[float, float, float] = (0.7, 0.15, 0.15),
        eval_all_splits: bool = False
    ):
        """Initialize LinearProbe.

        Args:
            dataset: Dataset containing sequences to be linear probed.
            embeddings: Embeddings of dataset sequences to be probed.
            task: Linear probing task. Valid tasks: {"regression", "reg_lin",
                "reg_ridge", "classification", "multilabel"}.
            target_col: Column from dataframe to use as labels.
            split_type: Method used for data split generation. Valid types:
                {"default", "homology"}.
            split_ratios: Ratio of data split sizes as a fraction of dataset.
            eval_all_splits: Evaluate metrics on all splits. Only evaluates
                validation split otherwise.
        """
        self.dataset = dataset
        self.data_df = dataset.data_df

        self.task = task
        self.target_col = target_col

        self.concat_embeddings(embeddings)

        self.splitter: DataSplitter

        if split_type == "homology":
            self.splitter = SPLIT_CATALOG[split_type](
                self.dataset.species
            )
        else:
            self.splitter = SPLIT_CATALOG["default"]()

        self.split_ratios = split_ratios
        self.eval_all_splits = eval_all_splits

    def concat_embeddings(self, embeddings: np.ndarray):
        """Merge embeddings with benchmark dataframe.

        Assumes that the dataframe rows and embedding order is identical.

        Args:
            embeddings: Embeddings in order of dataset. Shape (N x D).
        """
        self.data_df["embeddings"] = list(embeddings)

    def get_df_splits(
        self,
        random_seed: int,
    ) -> dict[str, np.ndarray]:
        """Get the train, validation, test data and label splits.

        Args:
            random_seed: Random seed used for generating data splits.

        Returns:
            Dictionary containing training, validation, testing splits for
            data and labels as numpy arrays.
        """
        train_df, val_df, test_df = self.splitter.get_all_splits_df(
            self.data_df,
            self.split_ratios,
            random_seed
        )

        splits = {
            "train_X": np.vstack(train_df["embeddings"]),
            "val_X": np.vstack(val_df["embeddings"]),
            "test_X": np.vstack(test_df["embeddings"]),
            "train_y": train_df[self.target_col],
            "val_y": val_df[self.target_col],
            "test_y": test_df[self.target_col],
        }

        if self.task == "multilabel":
            splits["train_y"] = np.vstack(splits["train_y"])
            splits["val_y"] = np.vstack(splits["val_y"])
            splits["test_y"] = np.vstack(splits["test_y"])

        return splits

    def run_linear_probe(self, random_seed: int = 2541) -> dict[str, float]:
        """Perform data split and run linear probe.

        Args:
            random_seed: Random seed used for data split.

        Returns:
            Dictionary of linear probing metrics per split.
        """
        try:
            model = self.linear_models[self.task]
        except KeyError:
            print("Invalid task name.")
            raise

        splits = self.get_df_splits(random_seed)

        np.random.seed(random_seed)
        model.fit(splits["train_X"], splits["train_y"])

        if self.task in ["regression", "reg_lin", "reg_ridge"]:
            metrics = self.eval_regression(model, splits)
        elif self.task == "classification":
            metrics = self.eval_classification(model, splits)
        elif self.task == "multilabel":
            metrics = self.eval_multilabel(model, splits)

        return metrics

    def eval_regression(
        self,
        model: RegressorMixin,
        splits: dict[str, np.ndarray]
    ) -> dict[str, float]:
        """Perform linear probing on regression task.

        Args:
            model: Scikit-learn regression model.
            splits: Data splits used for linear probing.

        Returns:
            Dictionary of linear probing metrics per split.
        """
        outputs = {"val": model.predict(splits["val_X"])}
        if self.eval_all_splits:
            outputs["train"] = model.predict(splits["train_X"])
            outputs["test"] = model.predict(splits["test_X"])

        metrics = {}

        for s_name, split_pred in outputs.items():
            split_y = splits[s_name + "_y"]
            metrics[s_name + "_mse"] = np.mean((split_pred - split_y) ** 2)
            metrics[s_name + "_r"] = pearsonr(split_pred, split_y).statistic

        return metrics

    def eval_classification(
        self,
        model: ClassifierMixin,
        splits: dict[str, np.ndarray]
    ) -> dict[str, float]:
        """Perform linear probing on classification task.

        Args:
            model: Scikit-learn classification model.
            splits: Data splits used for linear probing.

        Returns:
            Dictionary of linear probing metrics per split.
        """
        outputs = {"val": model.predict_proba(splits["val_X"])[:, 1]}
        if self.eval_all_splits:
            outputs["train"] = model.predict_proba(splits["train_X"])[:, 1]
            outputs["test"] = model.predict_proba(splits["test_X"])[:, 1]

        metrics = {}

        for s_name, split_pred in outputs.items():
            split_y = splits[s_name + "_y"]
            metrics[s_name + "_auroc"] = roc_auc_score(split_y, split_pred)
            metrics[s_name + "_auprc"] = average_precision_score(
                split_y,
                split_pred
            )

        return metrics

    def eval_multilabel(
        self,
        model: MultiOutputClassifier,
        splits: dict[str, np.ndarray]
    ) -> dict[str, float]:
        """Perform linear probing on multilabel classification task.

        Args:
            model: Scikit-learn classification model supporting multi-output.
            splits: Data splits used for linear probing.

        Returns:
            Dictionary of linear probing metrics per split.
        """
        outputs = {"val": model.predict_proba(splits["val_X"])}
        if self.eval_all_splits:
            outputs["train"] = model.predict_proba(splits["train_X"])
            outputs["test"] = model.predict_proba(splits["test_X"])

        metrics = {}

        for s_name, split_pred in outputs.items():
            split_pred = np.swapaxes(np.array(split_pred), 0, 1)[:, :, 1]

            split_y = splits[s_name + "_y"]
            metrics[s_name + "_auroc"] = roc_auc_score(
                split_y,
                split_pred,
                average="micro"
            )
            metrics[s_name + "_auprc"] = average_precision_score(
                split_y,
                split_pred,
                average="micro"
            )

        return metrics

    def linear_probe_multirun(
        self,
        random_seeds: list[int],
    ) -> dict[int, dict[str, float]]:
        """Run multiple linear probes with distinct data split randomization.

        Args:
            random_seeds: Trandom seeds used per individual linear probe run.

        Returns:
            Dictionary of metrics per random seed used to generate data splits
            for each individual linear probing run.
        """
        metrics = {}
        for random_seed in random_seeds:
            metric = self.run_linear_probe(random_seed)
            metrics[random_seed] = metric
        return metrics

    def compute_multirun_results(
        self,
        metrics: dict[int, dict[str, float]],
        print_output: bool = False,
        ci_multiplier: float = 1.96
    ) -> dict[str, str]:
        """Aggregate multi-run linear probing results.

        Prints mean metric and confidence interval at desired level.

        Args:
            metrics: Result of multi-run linear probing.
            print_output: Prints aggregated metrics.
            ci_multiplier: Constant multiplied to standard error to get CI.

        Returns:
            Dictionary of mean metric and CI per data split.
        """
        metric_vals: dict[str, list[float]] = {}

        for metric_dict in metrics.values():
            for metric_name, metric_val in metric_dict.items():
                metric_vals.setdefault(metric_name, []).append(metric_val)

        metric_mean = {k: np.mean(v) for k, v in metric_vals.items()}
        metric_std = {k: np.std(v) for k, v in metric_vals.items()}

        metric_out = {}

        for k in metric_vals.keys():
            se = ci_multiplier * (metric_std[k] / (np.sqrt(len(metrics))))
            metric_out[k] = "{} ± {}".format(metric_mean[k], se)

            if print_output:
                print("{} ± {}".format(metric_mean[k], se))

        return metric_out



from sklearn.model_selection import train_test_split

import pandas as pd

from mrna_bench.data_splitter.data_splitter import DataSplitter


class SklearnSplitter(DataSplitter):
    """Wrapper for sklearn train_test_split."""

    def split_df(
        self,
        df: pd.DataFrame,
        test_size: float,
        random_seed: int
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Split dataframe into train and test split using sklearn.

        Args:
            df: Dataframe to split.
            test_size: Fraction of dataset to place in test set.
            random_seed: Random seed used during sampling.

        Return:
            Train and test dataframes.
        """
        return train_test_split(
            df,
            test_size=test_size,
            random_state=random_seed
        )
